import torch,copy,os,random
from tqdm import trange
import numpy as np
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
import seaborn as sns
from dataclasses import dataclass, field
from .visualize import add_scale_bar

#################### Registration ####################
# Parameters class

@dataclass
class reg_params:
    dataname : str
    ### affine
    theta_r1 : float = 0
    theta_r2 : float = 0
    d_list : list[float] = field(default_factory=list)
    translation_params : list[float] = None
    mirror_t : list[float] = None
    alpha_basis : list[float] = field(default_factory=list)
    iterations : int = 500
    dist_penalty1 : float = 0
    attention_params: list[float] = field(default_factory=list)

    ### BS
    mesh_trans_list : list[float] = field(default_factory=list)
    attention_region : list[float] = field(default_factory=list)
    attention_params_bs : list[float] = field(default_factory=list)
    mesh_weight : list[float] = field(default_factory=list)
    iterations_bs : list[float] = field(default_factory=list)
    alpha_basis_bs : list[float] = field(default_factory=list)
    meshsize : list[float] = field(default_factory=list)
    img_size_bs : list[float] = field(default_factory=list) # max_xy
    dist_penalty2 : list[float] = field(default_factory=list)
    PaddingRate_bs : float = 0

    ### common
    bleeding : float = 500
    diff_step : float = 5
    min_qr2 : float = 0
    mean_q : float = 0
    mean_r : float = 0
    gpu: int = 0
    device : str = field(init=False)
    ifrigid : bool = False
    
    def __post_init__(self):
        if self.gpu != -1 and torch.cuda.is_available():
            self.device = 'cuda:{}'.format(self.gpu)
        else:
            self.device = 'cpu'

def get_range(sp_coords):
    yrng = max(sp_coords, key=lambda x:x[1])[1] - min(sp_coords, key=lambda x:x[1])[1]
    xrng = max(sp_coords, key=lambda x:x[0])[0] - min(sp_coords, key=lambda x:x[0])[0]
    return xrng, yrng

def prelocate(coords_q,coords_r,cov_anchor_it,bleeding,output_path,d_list=[1,2,3],prefix = 'test',ifplot = True,index_list = None,translation_params = None,mirror_t = None):
    idx_q = np.ones(coords_q.shape[0],dtype=bool) if index_list is None else index_list[0]
    idx_r = np.ones(coords_r.shape[0],dtype=bool) if index_list is None else index_list[1]
    mirror_t = [1,-1] if mirror_t is None else mirror_t
    theta_t = []
    J_t = []
    if translation_params is None:
        translation_x = [0]
        translation_y = [0]
    else:
        xrng, yrng = get_range(coords_r.detach().cpu())
        dx_ratio_max, dy_ratio_max, xy_steps = translation_params
        dx_max = dx_ratio_max * xrng
        dy_max = dy_ratio_max * yrng
        translation_x = np.linspace(-dx_max, dx_max, num=int(xy_steps)) # dx
        translation_y = np.linspace(-dy_max, dy_max, num=int(xy_steps)) # dy
    for mirror in mirror_t:
        for dx in translation_x:
            for dy in translation_y:
                for d in d_list:
                    for phi in [0,90,180,270]:
                        a = d
                        d = d * mirror
                        theta = torch.Tensor([a,d,phi,dx,dy]).reshape(5,1).to(coords_q.device)
                        coords_query_it = affine_trans_t(theta,coords_q)
                        try:
                            J_t.append(J_cal(coords_query_it[idx_q],coords_r[idx_r],cov_anchor_it,bleeding).sum().item())
                        except:
                            continue
                        theta_t.append(theta)
    if ifplot:
        prelocate_loss_plot(J_t,output_path,prefix)
    return(theta_t[np.argmin(J_t)])

def Affine_GD(coords_query_it_raw,coords_ref_it,cov_anchor_it,output_path,bleeding=500, dist_penalty = 0,diff_step = 50,alpha_basis = np.reshape(np.array([0,0,1/5,2,2]),[5,1]),iterations = 50,prefix='test',attention_params = [None,3,1,0],scale_t = 1,coords_log = False,index_list = None, mid_visual = False,early_stop_thres = 1, ifrigid = False):
    idx_q = np.ones(coords_query_it_raw.shape[0],dtype=bool) if index_list is None else index_list[0]
    idx_r = np.ones(coords_ref_it.shape[0],dtype=bool) if index_list is None else index_list[1]
    dev = coords_query_it_raw.device
    theta = torch.Tensor([1,1,0,0,0]).reshape(5,1).to(dev) # initial theta, [a,d,phi,t1,t2]
    coords_query_it = coords_query_it_raw.clone()
    plot_mid(coords_query_it.cpu() * scale_t,coords_ref_it.cpu() * scale_t,output_path,prefix + '_init',scale_bar_t=None) if mid_visual else None
    similarity_score = [J_cal(coords_query_it[idx_q],coords_ref_it[idx_r],cov_anchor_it,bleeding,dist_penalty,attention_params).sum().cpu().item()]
    it_J = []
    it_theta = []
    coords_q_log = []
    delta_similarity_score = [np.inf] * 5
    t = trange(iterations, desc='', leave=True)
    for it in t:
        alpha = alpha_init(alpha_basis,it,dev)
        ## de_sscore
        dJ_dxy_mat = dJ_dt_cal(coords_query_it[idx_q],
                      coords_ref_it[idx_r],
                      diff_step,
                      dev,
                      cov_anchor_it,
                      bleeding,
                      dist_penalty,
                      attention_params)

        dJ_dtheta = dJ_dtheta_cal(coords_query_it[idx_q,0],
                                  coords_query_it[idx_q,1],
                                  dJ_dxy_mat,theta,dev,ifrigid = ifrigid)
        theta = theta_renew(theta,dJ_dtheta,alpha,ifrigid = ifrigid)

        coords_query_it = affine_trans_t(theta,coords_query_it_raw)
        it_J.append(dJ_dtheta)
        it_theta.append(theta)
        if coords_log:
            coords_q_log.append(coords_query_it.detach().cpu().numpy())

        sscore_t = J_cal(coords_query_it[idx_q],coords_ref_it[idx_r],cov_anchor_it,bleeding,dist_penalty,attention_params).sum().cpu().item()
        # print(f'Loss: {sscore_t}')
        t.set_description(f'Loss: {sscore_t:.3f}')
        t.refresh()
        similarity_score.append(sscore_t)
        if mid_visual:
            if (it % 20 == 0) | (it == 0):
                plot_mid(coords_query_it.cpu() * scale_t,coords_ref_it.cpu() * scale_t,output_path,prefix + str(int(it/10 + 0.5)),scale_bar_t=None)
        if early_stop_thres is not None and it > 200:
            delta_similarity_score.append(similarity_score[-2] - similarity_score[-1])
            if np.all(np.array(delta_similarity_score[-5:]) < early_stop_thres):
                print(f'Early stop at {it}th iteration.')
                break
    return([similarity_score,it_J,it_theta,coords_q_log])

def BSpline_GD(coords_q,coords_r,cov_anchor_it,iterations,output_path,bleeding, dist_penalty = 0, alpha_basis = 1000,diff_step = 50,mesh_size = 5,prefix = 'test',mesh_weight = None,attention_params = [None,3,1,0],scale_t = 1,coords_log = False, index_list = None, mid_visual = False,max_xy = None,renew_mesh_trans = True,restriction_t = 0.5):
    idx_q = np.ones(coords_q.shape[0],dtype=bool) if index_list is None else index_list[0]
    idx_r = np.ones(coords_r.shape[0],dtype=bool) if index_list is None else index_list[1]
    dev = coords_q.device
    plot_mid(coords_q.cpu() * scale_t,coords_r.cpu()* scale_t,output_path,prefix + '_FFD_initial_' + str(iterations),scale_bar_t=None) if mid_visual else None

    max_xy = coords_q.max(0)[0].cpu() if max_xy is None else max_xy
    mesh,mesh_weight,kls,dxy_ffd_all,delta = BSpline_GD_preparation(max_xy,mesh_size,dev,mesh_weight)
    coords_query_it = coords_q.clone()

    similarity_score = [J_cal(coords_query_it[idx_q],coords_r[idx_r],cov_anchor_it,bleeding,dist_penalty,attention_params).sum().cpu().item()]
    mesh_trans_list = []
    coords_q_log = []
    mesh_trans = mesh.clone()
    max_movement = (max_xy / (mesh_size - 1.) * restriction_t).to(mesh.device).unsqueeze(-1).unsqueeze(-1)
    t = trange(iterations, desc='', leave=True)
    for it in t:
        dJ_dxy_mat = dJ_dt_cal(coords_query_it[idx_q],
                coords_r[idx_r],
                diff_step,
                dev,
                cov_anchor_it,
                bleeding,
                dist_penalty,
                attention_params)
        if renew_mesh_trans or it == 0:
            uv_raw, ij_raw = BSpline_GD_uv_ij_calculate(coords_query_it,delta,dev)
            uv = uv_raw[:,idx_q] # 2 * N[idx]
            ij = ij_raw[:,idx_q] # 2 * N[idx]

        result_B_t = B_matrix(uv,kls) ## 16 * N[idx]
        dxy_ffd = get_dxy_ffd(ij,result_B_t,mesh,dJ_dxy_mat,mesh_weight,alpha_basis)
        
        if renew_mesh_trans:
            mesh_trans = mesh + dxy_ffd
        else:
            mesh_trans = mesh + torch.clamp(mesh_trans + dxy_ffd - mesh, min=-max_movement, max=max_movement)
        mesh_trans_list.append(mesh_trans)
        coords_query_it = BSpline_renew_coords(uv_raw,kls,ij_raw,mesh_trans)
        if coords_log:
            coords_q_log.append(coords_query_it.detach().cpu().numpy())
        sscore_t = J_cal(coords_query_it[idx_q],coords_r[idx_r],cov_anchor_it,bleeding,dist_penalty,attention_params).sum().cpu().item()
        # print(f'Loss: {sscore_t}')
        t.set_description(f'Loss: {sscore_t:.3f}')
        t.refresh()

        similarity_score.append(sscore_t)
        if mid_visual:
            if (it % 20 == 0) | (it == 0):
                plot_mid(coords_query_it.cpu() * scale_t,coords_r.cpu() * scale_t,output_path,prefix + '_FFD_it_' + str(it),scale_bar_t=None)
                mesh_plot(mesh.cpu(),coords_q_t=coords_query_it.cpu(),mesh_trans_t=mesh_trans.cpu())
                plt.savefig(f'{output_path}/{prefix}_mesh_plot_it_{it}.pdf')
                plt.clf()
    ### visualization
    plt.figure(figsize=[20,10])
    plt.subplot(1,2,1)
    plt.scatter(np.array(coords_q.cpu()[:,0].tolist()) * scale_t,
        np.array(coords_q.cpu()[:,1].tolist()) * scale_t,  s=2,edgecolors='none', alpha = 0.5,rasterized=True,
        c='blue',label = 'Before')
    plt.scatter(np.array(coords_query_it.cpu()[:,0].tolist()) * scale_t,
        np.array(coords_query_it.cpu()[:,1].tolist()) * scale_t, s=2,edgecolors='none', alpha = 0.7,rasterized=True,
        c='#ef233c',label = 'After')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.legend(fontsize=15)
    plt.axis('equal')
    plt.subplot(1,2,2)
    titles = 'loss = ' + format(similarity_score[-1],'.1f')
    plt.scatter(list(range(0,len(similarity_score))),similarity_score,s = 5)
    plt.title(titles,fontsize=20)
    plt.savefig(os.path.join(output_path,prefix + '_after_Bspline_' + str(iterations) + '.pdf'))
    return([coords_query_it,mesh_trans_list,dxy_ffd_all,similarity_score,coords_q_log])

def J_cal(coords_q,coords_r,cov_mat,bleeding = 10, dist_penalty = 0,attention_params = [None,3,1,0]):
    attention_region,double_penalty,penalty_inc_all,penalty_inc_both = attention_params
    bleeding_x = coords_q[:, 0].min() - bleeding, coords_q[:, 0].max() + bleeding
    bleeding_y = coords_q[:, 1].min() - bleeding, coords_q[:, 1].max() + bleeding

    sub_ind = ((coords_r[:, 0] > bleeding_x[0]) & (coords_r[:, 0] < bleeding_x[1]) &
               (coords_r[:, 1] > bleeding_y[0]) & (coords_r[:, 1] < bleeding_y[1]))
    
    cov_mat_t = cov_mat[:,sub_ind]
    dist = torch.cdist(coords_q,coords_r[sub_ind,:])
    min_dist_values, close_idx = torch.min(dist, dim=1)

    tmp1 = torch.stack((torch.arange(coords_q.shape[0], device=coords_q.device), close_idx)).T
    s_score_mat = cov_mat_t[tmp1[:, 0], tmp1[:, 1]]
    
    if(dist_penalty != 0):
        penalty_tres = torch.sqrt((coords_r[:,0].max() - coords_r[:,0].min()) * (coords_r[:,1].max() - coords_r[:,1].min()) / coords_r.shape[0])
        dist_d = min_dist_values / penalty_tres
        if(type(attention_region) is np.ndarray):
            attention_region = torch.tensor(attention_region, device=coords_q.device)
            dist_d[attention_region] = min_dist_values[attention_region] / (penalty_tres/double_penalty)
            dist_d[dist_d < 1] = 1
            dist_d[dist_d > 1] *= dist_penalty
            dist_d[attention_region] *= penalty_inc_all
            dist_d[(dist_d > 1) & attention_region] *= (penalty_inc_both/dist_penalty + 1)
        else:
            dist_d[dist_d < 1] = 1
            dist_d[dist_d > 1] *= dist_penalty
        return s_score_mat * dist_d
    return s_score_mat

def alpha_init(alpha_basis,it,dev):
    return 5/torch.pow(torch.Tensor([it/40 + 1]).to(dev),0.6) * alpha_basis

def dJ_dt_cal(coords_q,coords_r,diff_step,dev,cov_anchor_it,bleeding,dist_penalty,attention_params):
    dJ_dy = (J_cal(coords_q + torch.tensor([0,diff_step], device=dev),
                          coords_r,
                          cov_anchor_it,
                          bleeding,
                          dist_penalty,
                          attention_params) - 
               J_cal(coords_q + torch.tensor([0,-diff_step], device=dev),
                          coords_r,
                          cov_anchor_it,
                          bleeding,
                          dist_penalty,
                          attention_params)) / (2 * diff_step)
    dJ_dx = (J_cal(coords_q + torch.tensor([diff_step,0], device=dev),
                          coords_r,
                          cov_anchor_it,
                          bleeding,
                          dist_penalty,
                          attention_params) - 
               J_cal(coords_q + torch.tensor([-diff_step,0], device=dev),
                          coords_r,
                          cov_anchor_it,
                          bleeding,
                          dist_penalty,
                          attention_params)) / (2 * diff_step)
    dJ_dxy_mat = torch.vstack((dJ_dx,dJ_dy)) # [dJ_{i}/dx_{i},dJ_{i}/dy_{i}] (2 * N)
    return dJ_dxy_mat

def dJ_dtheta_cal(xi,yi,dJ_dxy_mat,theta,dev,ifrigid = False):
    '''
    #dxy_da:
    #{x * cos(rad_phi), x * sin(rad_phi)}
    #dxy_dd:
    #{-y * sin(rad_phi), y * cos(rad_phi)}
    #dxy_dphi:
    #{-d * y * cos(rad_phi) - a * x * sin(rad_phi), a * x * cos(rad_phi) - d * y * sin(rad_phi)}
    #dxy_dt1:
    #{1, 0}
    #dxy_dt2:
    #{0, 1}

    # when we set d = a (rigid):
    #dxy_da 
    #{x * cos(rad_phi) - y * sin(rad_phi), y * cos(rad_phi) + x * sin(rad_phi)}
    #dxy_dd - set as the same value as dxy_da
    #{x * cos(rad_phi) - y * sin(rad_phi), y * cos(rad_phi) + x * sin(rad_phi)}
    #dxy_dphi
    #{-a * y * cos(rad_phi) - a * x * sin(rad_phi), a * x * cos(rad_phi) - a * y * sin(rad_phi)}
    '''
    N = xi.shape[0]
    rad_phi = theta[2,0].deg2rad()
    cos_rad_phi = rad_phi.cos()
    sin_rad_phi = rad_phi.sin()
    ones = torch.ones(N, device=dev)
    zeros = torch.zeros(N, device=dev)
    if ifrigid:
        #### let d = a, only allow scaling, rotation and translation (Similarity transformation)
        #### If we want to use pure rigid transformation, just set `alpha_basis` as `[0,0,x,x,x]`, then the theta[0] will be always 1.
        dxy_dtheta = torch.stack([
            torch.stack([
                xi * cos_rad_phi - yi * sin_rad_phi, #dxy_da (rigid)
                xi * cos_rad_phi - yi * sin_rad_phi, #dxy_dd - won't use (rigid)
                -theta[0] * cos_rad_phi * yi - theta[0] * xi * sin_rad_phi, #dxy_dphi
                ones, #dxy_dt1
                zeros]), #dxy_dt2
            torch.stack([
                yi * cos_rad_phi + xi * sin_rad_phi, #dxy_da (rigid)
                yi * cos_rad_phi + xi * sin_rad_phi, #dxy_dd - won't use (rigid)
                theta[0] * xi * cos_rad_phi - theta[0] * yi * sin_rad_phi, #dxy_dphi
                zeros, #dxy_dt1
                ones])]) #dxy_dt2
    else:
        dxy_dtheta = torch.stack([
            torch.stack([
                xi * cos_rad_phi, #dxy_da
                -yi * sin_rad_phi, #dxy_dd
                -theta[1] * cos_rad_phi * yi - theta[0] * xi * sin_rad_phi, #dxy_dphi
                ones, #dxy_dt1
                zeros]), #dxy_dt2
            torch.stack([
                xi * sin_rad_phi, #dxy_da
                yi * cos_rad_phi, #dxy_dd
                theta[0] * xi * cos_rad_phi - theta[1] * yi * sin_rad_phi, #dxy_dphi
                zeros, #dxy_dt1
                ones])]) #dxy_dt2

    dJ_dtheta = torch.bmm(dxy_dtheta.permute(2, 1, 0), ### [N,5,2]
                          dJ_dxy_mat.transpose(0, 1).unsqueeze(-1) ### [N,2,1]
                          ).squeeze(2) # [dJ_{i}/dtheta_{k}] (N * 5)
    dJ_dtheta = dJ_dtheta.sum(0)

    return dJ_dtheta

def theta_renew(theta,dJ_dtheta,alpha,ifrigid = False):
    alpha_dJ = alpha * dJ_dtheta.reshape(5,1)
    alpha_dJ[0:3] = alpha_dJ[0:3] / 1000 # avoid dtheta_{abcd} change a lot of x and y
    if ifrigid & (theta[0] == -theta[1]):
        # only when the rigid transformation is allowed, we should check the value of d and a if they are mirrored.
        # if d and a are mirrored (setting in the prelocate `d = d * mirror``), we should set alpha_dJ[1] as the `-alpha_dJ[1]`.
        alpha_dJ[1] = -alpha_dJ[1] 
    theta_new = theta - alpha_dJ
    return theta_new

def affine_trans_t(theta,coords_t):
    rad_phi = theta[2,0].deg2rad()
    cos_rad_phi = rad_phi.cos()
    sin_rad_phi = rad_phi.sin()
    A = torch.Tensor([[theta[0,0] * cos_rad_phi, -theta[1,0] * sin_rad_phi],[theta[0,0] * sin_rad_phi, theta[1,0] * cos_rad_phi]]).to(theta.device)
    t_vec = theta[3:5,:]
    coords_t1 = torch.mm(A,coords_t.T) + t_vec
    coords_t1 = coords_t1.T
    return coords_t1

def torch_Bspline(uv, kl):
    return (
        torch.where(kl == 0, (1 - uv) ** 3 / 6,
        torch.where(kl == 1, uv ** 3 / 2 - uv ** 2 + 2 / 3,
        torch.where(kl == 2, (-3 * uv ** 3 + 3 * uv ** 2 + 3 * uv + 1) / 6,
        torch.where(kl == 3, uv ** 3 / 6, torch.zeros_like(uv)))))
    )

def BSpline_GD_preparation(max_xy,mesh_size,dev,mesh_weight):
    delta = max_xy / (mesh_size - 1.)
    mesh = np.ones((2, mesh_size + 3, mesh_size + 3)) ## 2 * (mesh_size + 3) * (mesh_size + 3)
    for i in range(mesh_size + 3):
        for j in range(mesh_size + 3):
            mesh[:, i, j] = [(i - 1) * delta[0], (j - 1) * delta[1]] ## 0 - -delta, 1 - 0, 2 - delta, ..., 6 - delta * 5, 7 - delta * 6 (last row)
    mesh = torch.tensor(mesh).to(dev)
    mesh_weight = torch.tensor(mesh_weight).to(dev) if type(mesh_weight) is np.ndarray else 1
    kls = torch.stack(torch.meshgrid(torch.arange(4), torch.arange(4))).flatten(1).to(dev) ## 2 * 16
    dxy_ffd_all = torch.zeros(mesh.shape, device=dev) ## 2 * (mesh_size + 3) * (mesh_size + 3)
    return mesh,mesh_weight,kls,dxy_ffd_all,delta

def BSpline_GD_uv_ij_calculate(coords_query_it,delta,dev):
    pos_reg = coords_query_it.T / delta.reshape(2,1).to(dev) # 2 * N
    pos_floor = pos_reg.floor().long() # 2 * N
    uv_raw = pos_reg - pos_floor # 2 * N
    ij_raw = pos_floor - 1 # 2 * N
    return uv_raw, ij_raw

def B_matrix(uv_t, kls_t):
    result_B_list = []
    for kl in kls_t.T:
        B = torch_Bspline(uv_t, kl.view(2, 1)) # 2 * N[idx]
        result_B_list.append(B.prod(0, keepdim=True)) # 1 * N[idx] ; .prod() - product of all elements in the tensor along a given dimension (0 - reduce along rows, 1 - reduce along columns)
    return torch.cat(result_B_list,0) # 16 * N[idx]

def get_dxy_ffd(ij,result_B_t,mesh,dJ_dxy_mat,mesh_weight,alpha_basis):
    dxy_ffd_t = torch.zeros(mesh.shape, device=result_B_t.device)
    ij_0 = ij[0] + 1
    ij_1 = ij[1] + 1
    for k in range(dxy_ffd_t.shape[1]):
        for l in range(dxy_ffd_t.shape[2]):
            mask = (ij_0 <= k) & (k <= ij_0 + 3) & (ij_1 <= l) & (l <= ij_1 + 3)
            if mask.any():  # check if there is any True in the mask
                idx_kl = mask.nonzero().flatten()
                ij_t = torch.tensor([k, l], device=ij.device) - (ij[:, idx_kl].T + 1)
                keys = ij_t[:, 0] * 4 + ij_t[:, 1]
                t33 = result_B_t[keys, idx_kl]
                dxy_ffd_t[:,k,l] -= torch.matmul(dJ_dxy_mat[:,idx_kl],t33.unsqueeze(1).float()).squeeze(1)
    dxy_ffd_t *= mesh_weight
    dxy_ffd_t = dxy_ffd_t * alpha_basis
    return dxy_ffd_t

def BSpline_renew_coords(uv_t,kls_t,ij_t,mesh_trans):
    result_tt = torch.zeros_like(uv_t, dtype=torch.float32)
    for kl in kls_t.T:
        B = torch_Bspline(uv_t, kl.view(2, 1))
        pivots = (ij_t + 1 + kl.view(2, 1)).clamp(0, mesh_trans.size(-1) - 1)
        mesh_t = mesh_trans[:, pivots[0], pivots[1]]
        result_tt += B.prod(0, keepdim=True) * mesh_t
    return result_tt.T

def reg_total_t(coords_q,coords_r,params_dist):
    dev = params_dist.device
    mean_q = coords_q.mean(0)
    mean_r = coords_r.mean(0)
    coords_q_t = torch.tensor(np.array(coords_q) - mean_q).float().to(dev) ## Initial location
    coords_q_r1 = affine_trans_t(params_dist.theta_r1,coords_q_t) ## Prelocation 1st Affine
    coords_q_r2 = affine_trans_t(params_dist.theta_r2,coords_q_r1) ## Affine transformation 2st Affine
    if params_dist.mesh_trans_list != [] and params_dist.mesh_trans_list != [[]]:
        coords_q_r3 = coords_q_r2.clone()
        for round_t in range(len(params_dist.mesh_trans_list)):
            coords_q_r3 = coords_q_r3.clone() - params_dist.min_qr2[round_t]
            coords_q_r3 = FFD_Bspline_apply_t(coords_q_r3.clone(),params_dist,round_t)
            coords_q_r3 = coords_q_r3.clone() + params_dist.min_qr2[round_t]
        coords_q_f = coords_q_r3.clone()
    else:
        coords_q_f = coords_q_r2
    coords_q_reconstruct = coords_q_f + torch.tensor(mean_r).to(dev)
    coords_q_reconstruct = coords_q_reconstruct.float()
    return coords_q_f,coords_q_reconstruct

def FFD_Bspline_apply_t(coords_q,params_dist,round_t = 0):
    mesh_trans_list = params_dist.mesh_trans_list[round_t]
    dev = coords_q.device
    img_size = params_dist.img_size_bs[round_t]
    mesh_size = mesh_trans_list[0].shape[2] - 3
    delta = img_size / (mesh_size - 1.)
    coords_query_it = copy.deepcopy(coords_q)

    for it in trange(len(mesh_trans_list), desc='', leave=True):
        mesh_trans = mesh_trans_list[it]
        pos_reg = coords_query_it.T / delta.reshape(2,1).to(dev)
        pos_floor = pos_reg.floor().long()
        uv = pos_reg - pos_floor
        ij = pos_floor - 1
        kls = torch.stack(torch.meshgrid(torch.arange(4), torch.arange(4))).flatten(1).to(dev)
        result_tt = torch.zeros_like(uv).float()
        for kl in kls.T:
            B = torch_Bspline(uv, kl.view(2, 1))
            pivots = (ij + 1 + kl.view(2, 1)).clamp(0, mesh_trans.size(-1) - 1)
            mesh_t = mesh_trans[:, pivots[0], pivots[1]]
            result_tt += B.prod(0, keepdim=True) * mesh_t
        coords_query_it = result_tt.T
    return coords_query_it

def rescale_coords(coords_raw,graph_list,rescale = False):
    rescale_factor = 1
    if rescale:
        coords_raw = coords_raw.copy()
        for sample_t in graph_list:
            rescale_factor_t = 22340 / np.abs(coords_raw[sample_t]).max()
            coords_raw[sample_t] = coords_raw[sample_t].copy() * rescale_factor_t
            if sample_t == graph_list[1]:
                rescale_factor = rescale_factor_t
    return coords_raw,rescale_factor

#################### Visualization ####################

def mesh_plot(mesh_t,coords_q_t,mesh_trans_t = None):
    mesh_no_last_row = mesh_t[:, :, :].numpy()
    plt.figure(figsize=[10,10])
    plt.plot(mesh_no_last_row[0], mesh_no_last_row[1], 'blue')
    plt.plot(mesh_no_last_row.T[..., 0], mesh_no_last_row.T[..., 1], 'blue')
    if(type(mesh_trans_t) is not type(None)):
        mesh_trans_no_last_row = mesh_trans_t[:, :, :].numpy()
        plt.plot(mesh_trans_no_last_row[0], mesh_trans_no_last_row[1], 'orange')
        plt.plot(mesh_trans_no_last_row.T[..., 0], mesh_trans_no_last_row.T[..., 1], 'orange')
    plt.scatter(coords_q_t.T[0,:],coords_q_t.T[1,:],c='blue',s = 0.5,alpha=0.5, rasterized=True)

def plot_mid(coords_q,coords_r,output_path='',filename = None,title_t = ['ref','query'],s_t = 8,scale_bar_t = None):
    plt.rcParams.update({'font.size' : 30,'axes.titlesize' : 30,'pdf.fonttype':42,'legend.markerscale' : 5})
    plt.figure(figsize=[10,12])
    plt.scatter(np.array(coords_r)[:,0].tolist(),
        np.array(coords_r)[:,1].tolist(),  s=s_t,edgecolors='none', alpha = 0.5,rasterized=True,
        c='#9295CA',label = title_t[0])
    plt.scatter(np.array(coords_q)[:,0].tolist(),
        np.array(coords_q)[:,1].tolist(), s=s_t,edgecolors='none', alpha = 0.5,rasterized=True,
        c='#E66665',label = title_t[1])
    plt.legend(fontsize=15)
    plt.axis('equal')
    if (type(scale_bar_t) != type(None)):
        add_scale_bar(scale_bar_t[0],scale_bar_t[1])
    if (filename != None):
        plt.savefig(os.path.join(output_path,filename + '.pdf'),dpi = 100)

def corr_heat(coords_q,coords_r,corr,output_path,title_t = ['Corr in ref','Anchor in query'],filename=None,scale_bar_t = None):
    plt.rcParams.update({'font.size' : 20,'axes.titlesize' : 20,'pdf.fonttype':42})
    random.seed(2)
    sampled_points = np.sort(random.sample(list(range(0,coords_q.shape[0])),20))
    plt.figure(figsize=((40,25)))
    for t in range(0,len(sampled_points)):

        plt_ind = t * 2
        ins_cell_idx = sampled_points[t]
        col_value = corr[ins_cell_idx,:]
        col_value_bg = [0] * coords_q.shape[0]
        col_value_bg[ins_cell_idx] = 1
        size_value_bg = [5] * coords_q.shape[0]
        size_value_bg[ins_cell_idx] = 30
        plt.subplot(5,8,plt_ind + 1)
        plt.scatter(np.array(coords_r[:,0]), np.array(coords_r[:,1]), s=5,edgecolors='none',
            c=col_value,cmap = 'vlag',vmin = -1,vmax= 1,rasterized=True)

        plt.title(title_t[0])
        plt.axis('equal')
        if (type(scale_bar_t) != type(None)):
            add_scale_bar(scale_bar_t[0],scale_bar_t[1])
        plt.subplot(5,8,plt_ind + 2)
        plt.scatter(np.array(coords_q[:,0]), np.array(coords_q[:,1]), s=size_value_bg,edgecolors='none',
            c=col_value_bg,cmap = 'vlag',vmin = -1,vmax= 1,rasterized=True)
        plt.scatter(np.array(coords_q[ins_cell_idx,0]), np.array(coords_q[ins_cell_idx,1]), s=size_value_bg[ins_cell_idx],edgecolors='none',
            c=col_value_bg[ins_cell_idx],cmap = 'vlag',vmin = -1,vmax= 1,rasterized=True)
        plt.title(title_t[1])
        plt.axis('equal')
        if (type(scale_bar_t) != type(None)):
            add_scale_bar(scale_bar_t[0],scale_bar_t[1])
    plt.tight_layout()
    plt.colorbar()
    if (filename != None):
        plt.savefig(os.path.join(output_path,filename + '.pdf'),dpi=100,transparent=True)

def prelocate_loss_plot(J_t,output_path,prefix = 'test'):
    plt.rcParams.update({'font.size' : 15})
    plt.figure(figsize=[5,5])
    plt.scatter(x=list(range(0,len(J_t))),y=J_t)
    plt.savefig(f'{output_path}/{prefix}_prelocate_loss.pdf')

def register_result(coords_q,coords_r,cov_anchor_t,bleeding,embed_stack,output_path,k=8,prefix='test',scale_t = 1,index_list = None):
    idx_q = np.ones(coords_q.shape[0],dtype=bool) if index_list is None else index_list[0]
    idx_r = np.ones(coords_r.shape[0],dtype=bool) if index_list is None else index_list[1]
    coords_q = coords_q * scale_t
    coords_r = coords_r * scale_t
    kmeans = KMeans(n_clusters=k,random_state=0).fit(embed_stack)
    cell_label = kmeans.labels_
    cluster_pl = sns.color_palette('tab20',len(np.unique(cell_label)))
    ### panel 1 ###
    plot_mid(coords_q[idx_q],coords_r[idx_r],output_path,f'{prefix}_Results_1', scale_bar_t = None)
    ### panel 2 ###
    plt.figure(figsize=[10,12])
    plt.rcParams.update({'font.size' : 10,'axes.titlesize' : 20,'pdf.fonttype':42})
    col=coords_q[idx_q,0]
    row=coords_q[idx_q,1]
    cell_type_t = cell_label[0:coords_q[idx_q].shape[0]]
    for i in set(cell_type_t):
        plt.scatter(np.array(col)[cell_type_t == i],
        np.array(row)[cell_type_t == i], s=12,edgecolors='none',alpha = 0.5,rasterized=True,
        c=np.array(cluster_pl)[cell_type_t[cell_type_t == i]],label = str(i))
    col=coords_r[idx_r,0]
    row=coords_r[idx_r,1]
    cell_type_t = cell_label[coords_q[idx_q].shape[0]:]
    for i in set(cell_type_t):
        plt.scatter(np.array(col)[cell_type_t == i],
        np.array(row)[cell_type_t == i], s=12,edgecolors='none',alpha = 0.5,rasterized=True,
        c=np.array(cluster_pl)[cell_type_t[cell_type_t == i]],label = str(i))
    plt.axis('equal')
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.title('K means (k = ' + str(k) + ')',fontsize=30)
    add_scale_bar(200,'200 µm')
    plt.savefig(f'{output_path}/{prefix}_Results_2.pdf',dpi = 300)
    ### panel 3 ###
    plt.figure(figsize=[20,12])
    plt.subplot(1,2,1)
    t_score = J_cal(torch.from_numpy(coords_q[idx_q]),torch.from_numpy(coords_r[idx_r]),cov_anchor_t,bleeding)
    plt.scatter(coords_q[idx_q,0],coords_q[idx_q,1],c=1 - t_score,cmap = 'vlag',vmin = -1,vmax = 1,s = 15,edgecolors='none',alpha=0.5,rasterized=True)
    add_scale_bar(200,'200 µm')
    plt.subplot(1,2,2)
    plt.scatter(coords_q[0,0],coords_q[0,1],c=[0],cmap = 'vlag',vmin = -1,vmax = 1,s = 15,alpha=0.5)
    plt.axis('off')
    plt.colorbar()
    plt.savefig(f'{output_path}/{prefix}_Results_3.pdf',dpi = 300)

def affine_reg_params(it_theta,similarity_score,iterations,output_path,prefix='test'):
    plt.rcParams.update({'font.size' : 15,'axes.titlesize' : 15,'pdf.fonttype':42})
    similarity_score_t = copy.deepcopy(similarity_score)
    titles = ['a','d','φ','t1','t2','loss = ' + format(similarity_score[-1],'.1f')]
    plt.figure(figsize=[15,8])
    for i in range(0,6):
        plt.subplot(2,4,i+1)
        if i == 5:
            plt.scatter(list(range(0,len(similarity_score_t))),similarity_score_t,s = 5)
        else:
            # plt.scatter(x = range(0,iterations),y=np.array(it_theta)[:,i,0],s = 5)
            plt.scatter(x = range(1,len(similarity_score_t)),y=np.array(it_theta)[:,i,0],s = 5)
        plt.title(titles[i],fontsize=20)
    plt.savefig(os.path.join(output_path,prefix + '_params_Affine_GD_' + str(iterations) + 'its.pdf'))

def CAST_STACK_rough(coords_raw_list, ifsquare=True, if_max_xy=True, percentile = None):
    '''
    coords_raw_list: list of numpy arrays, each array is the coordinates of a layer
    ifsquare: if True, the coordinates will be scaled to a square
    if_max_xy: if True, the coordinates will be scaled to the max value of the `max_range_x` and `max_range_y`, respectively (if ifsquare is False), or the max value of [max_range_x,max_range_y] (if ifsquare is True)
    percentile: if not None, the min and max will be calculated based on the percentile of the coordinates for each slice.
    '''
    # Convert list of arrays to a single numpy array for easier processing
    all_coords = np.concatenate(coords_raw_list)
    # Finding the global min and max for both x and y
    if percentile is None:
        min_x, min_y = np.min(all_coords, axis=0)
        max_x, max_y = np.max(all_coords, axis=0)
    else:
        min_x_list, min_y_list, max_x_list, max_y_list = [], [], [], []
        for coords_t in coords_raw_list:
            min_x_list.append(np.percentile(coords_t[:,0],percentile))
            min_y_list.append(np.percentile(coords_t[:,1],percentile))
            max_x_list.append(np.percentile(coords_t[:,0],100-percentile))
            max_y_list.append(np.percentile(coords_t[:,1],100-percentile))
        min_x, min_y = np.min(min_x_list), np.min(min_y_list)
        max_x, max_y = np.max(max_x_list), np.max(max_y_list)
    max_xy = np.array([max_x - min_x, max_y - min_y])
    scaled_coords_list = []
    for coords_t in coords_raw_list:
        coords_t2 = (coords_t - coords_t.min(axis=0)) / np.ptp(coords_t, axis=0)
        if if_max_xy:
            max_xy_scale = max_xy
        else:
            max_xy_scale = max_xy / np.max(max_xy)
        scaled_coords = coords_t2 * np.max(max_xy_scale) if ifsquare else coords_t2 * max_xy_scale
        scaled_coords_list.append(scaled_coords)
    return scaled_coords_list

#################### Calculation ####################
def coords_minus_mean(coord_t):
    return np.array(coord_t) - np.mean(np.array(coord_t),axis = 0)

def coords_minus_min(coord_t):
    return np.array(coord_t) - np.min(np.array(coord_t),axis = 0)

def max_minus_value(corr):
    return np.max(corr) - corr

def coords_minus_min_t(coord_t):
    return coord_t - coord_t.min(0)[0]

def max_minus_value_t(corr):
    return corr.max() - corr

def corr_dist(query_np, ref_np, nan_as = 'min'):
    from sklearn.metrics import pairwise_distances_chunked
    def chunked_callback(dist_matrix,start):
        return 1 - dist_matrix
    chunks = pairwise_distances_chunked(query_np, ref_np, metric='correlation', n_jobs=-1, working_memory=1024, reduce_func=chunked_callback)
    corr_q_r = np.vstack(list(chunks))
    if nan_as == 'min':
        corr_q_r[np.isnan(corr_q_r)] = np.nanmin(corr_q_r)
    return corr_q_r

def region_detect(embed_dict_t,coords0,k = 20):
    plot_row = int(np.floor((k+1)/4) + 1)
    kmeans = KMeans(n_clusters=k,random_state=0).fit(embed_dict_t)
    cell_label = kmeans.labels_
    cluster_pl = sns.color_palette('tab20',len(np.unique(cell_label)))
    plt.figure(figsize=((20,5 * plot_row)))
    plt.subplot(plot_row,4,1)
    cell_label_idx = 0
    col=coords0[:,0].tolist()
    row=coords0[:,1].tolist()
    cell_type_t = cell_label[cell_label_idx:(cell_label_idx + coords0.shape[0])]
    cell_label_idx += coords0.shape[0]
    for i in set(cell_type_t):
        plt.scatter(np.array(col)[cell_type_t == i],
        np.array(row)[cell_type_t == i], s=5,edgecolors='none',
        c=np.array(cluster_pl)[cell_type_t[cell_type_t == i]],label = str(i))
    plt.title(' (KMeans, k = ' + str(k) + ')',fontsize=20)
    plt.xticks(fontsize=20)
    plt.yticks(fontsize=20)
    plt.axis('equal')
    for j,i in enumerate(set(cell_type_t)):
        plt.subplot(plot_row,4,j+2)
        plt.scatter(np.array(col),np.array(row),s=3,c = '#DDDDDD')
        plt.scatter(np.array(col)[cell_type_t == i],
        np.array(row)[cell_type_t == i], s=5,edgecolors='none',
        c=np.array(cluster_pl)[cell_type_t[cell_type_t == i]],label = str(i))
        plt.title(str(i),fontsize=20)
        plt.axis('equal')
    return cell_label



from typing import Optional

import os
from pathlib import Path

import pandas as pd
import numpy as np
from itertools import permutations, product
import scanpy as sc
import requests
from tqdm.auto import tqdm


zipfile_install = False

def check_zipfile():
    global zipfile_install
    try:
        import zipfile
        zipfile_install=True
    except ImportError:
        raise ImportError(
            'Please install the zipfile: `pip install zipfile`.'
        )
    

def _download_from_url(file_url: str, save_path: Path):
    check_zipfile()
    global zipfile_install
    if zipfile_install==True:
        global zipfile
        import zipfile


    try:
        response = requests.get(file_url, stream=True)
        response.raise_for_status()
    except requests.exceptions.RequestException as e:
        print(f'Download error: {e}')
        return

    total_size = int(response.headers.get('content-length', 0))
    block_size = 8192
    progress_bar = tqdm(total=total_size, unit='B', unit_scale=True)

    download_file = save_path.parent / file_url.split('/')[-1]
    with open(download_file, 'wb') as file:
        for chunk in response.iter_content(chunk_size=block_size):
            if chunk:
                file.write(chunk)
                progress_bar.update(len(chunk))

    progress_bar.close()

    if str(download_file).endswith('.zip'):
        with zipfile.ZipFile(download_file, 'r') as zip_file:
            zip_file.extractall(os.path.dirname(save_path))
            zip_file.close()
        os.remove(download_file)

    print(f'Ths data has been downloaded to `{save_path}`.')


def load_human_prior_interaction_network(dataset: str = 'nichenet',
                                         only_directed: bool = False,
                                         force_download: bool = False):

    # The URL for every dataset. These datasets are stored at zenodo (https://doi.org/10.5281/zenodo.7564872).
    urls = {
        'nichenet': 'https://zenodo.org/record/8013900/files/NicheNet_human.zip',
        'pathwaycommons': 'https://zenodo.org/record/8013900/files/PathwayCommons12.All.hgnc.zip',
        'inbiomap': 'https://zenodo.org/record/8013900/files/InBioMap.zip',
        'harmonizome': 'https://zenodo.org/record/8013900/files/Harmonizome_nichenet.zip',
        'omnipath_interactions': 'https://zenodo.org/record/8013900/files/Omnipath_interaction.zip',
    }
    filenames = {
        'nichenet': 'NicheNet_human.csv',
        'pathwaycommons': 'PathwayCommons12.All.hgnc.sif',
        'inbiomap': 'InBioMap.csv',
        'harmonizome': 'Harmonizome_nichenet.csv',
        'omnipath_interactions': 'Omnipath_interaction.csv',
    }

    # Download if the file does not exist
    data_path = Path('./data_cache') / filenames[dataset]
    if force_download or not data_path.exists():
        data_path.parent.mkdir(parents=True, exist_ok=True)
        _download_from_url(urls[dataset], data_path)

    if dataset == 'nichenet':  # 5,583,023
        prior_net = pd.read_csv(data_path, index_col=None, header=0)

    elif dataset == 'pathwaycommons':  # 1,200,159
        prior_net = pd.read_csv(data_path, sep='\t',
                                names=['from', 'type', 'to'])
        undirected_type = ['interacts-with', 'in-complex-with']
        type_mapper = dict({v: 1 for v in list(set(prior_net.type.unique()) - set(undirected_type))},
                           **{v: 0 for v in undirected_type})
        prior_net['is_directed'] = prior_net['type'].map(type_mapper)
        prior_net = prior_net.loc[~prior_net['from'].str.startswith('CHEBI'), :]
        prior_net = prior_net.loc[~prior_net['to'].str.startswith('CHEBI'), :]
        del prior_net['type']

    elif dataset == 'inbiomap':  # 625,641
        prior_net = pd.read_csv(data_path)
        prior_net.rename(columns={'genesymbol_a': 'from', 'genesymbol_b': 'to'}, inplace=True)
        prior_net = prior_net.loc[:, ['from', 'to']]

    elif dataset == 'harmonizome':  # 3,418,949
        prior_net = pd.read_csv(data_path, sep='\t')
        prior_net = prior_net.loc[:, ['from', 'to']]

    elif dataset == 'omnipath_interactions':  # 525,430
        prior_net = pd.read_csv(data_path)
        prior_net.rename(columns={'source_genesymbol': 'from', 'target_genesymbol': 'to'}, inplace=True)
        complex_idx = prior_net['source'].str.startswith('COMPLEX') | \
                      prior_net['target'].str.startswith('COMPLEX')
        prior_net_genes_only = prior_net.loc[~complex_idx, ['from', 'to', 'is_directed']]
        prior_net_genes_complex = prior_net.loc[complex_idx, ['from', 'to', 'is_directed']]
        # process complex items
        temp_edge = []
        temp_edge_type = []
        for source, target in zip(prior_net_genes_complex['from'], prior_net_genes_complex['to']):
            source = source.split('_')
            target = target.split('_')
            # inside the complex
            intra = list(permutations(source, r=2)) + list(permutations(target, r=2))
            temp_edge += intra
            temp_edge_type += [0] * len(intra)
            # between two complexes
            inter = list(product(source, target))
            temp_edge += inter
            temp_edge_type += [1] * len(inter)
        prior_net_complex = pd.DataFrame(temp_edge, columns=['from', 'to'])
        prior_net_complex['is_directed'] = temp_edge_type
        prior_net = pd.concat([prior_net_genes_only, prior_net_complex], axis=0)
        prior_net = prior_net.dropna()
        prior_net = prior_net.drop_duplicates()

    else:
        print(f"Value error. {dataset} is not available.")
        print("Available option: {'nichenet', 'pathwaycommons', 'inbiomap', 'harmonizome', 'omnipath_interactions'}")

    if ('is_directed' in prior_net) and only_directed:
        prior_net = prior_net[prior_net['is_directed'] == 1]

    prior_net = prior_net[['from', 'to']].drop_duplicates().astype(str)
    print(f"Load the prior gene interaction network: {dataset}. "
          f"#Genes: {len(np.unique(prior_net.iloc[:, [0, 1]]))}, #Edges: {len(prior_net)}")

    return prior_net


def convert_human_to_mouse_network(net: pd.DataFrame):
    import biomart

    print('Convert genes of the prior interaction network to mouse gene symbols:')
    with tqdm(total=10, desc='Processing', miniters=1) as outer_bar:
        outer_bar.update()

        # Set up connection to server
        for name in ['ensembldb', 'asia', 'useast', 'martdb']:
            try:
                server = biomart.BiomartServer(
                    f'http://{name}.ensembl.org/biomart/')
                print(f'Server \'http://{name}.ensembl.org/biomart/\' is OK')
                break
            except Exception as e:
                print(f'404 Client Error: Not Found for url: http://{name}.ensembl.org/biomart//martservice')

        human_dataset = server.datasets['hsapiens_gene_ensembl']
        outer_bar.update()
        mouse_dataset = server.datasets['mmusculus_gene_ensembl']
        outer_bar.update()

        human_attributes = ['ensembl_gene_id', 'hgnc_symbol']
        mouse_attributes = ['ensembl_gene_id', 'mgi_symbol']  # 'external_gene_name'
        to_homolog_attribute = 'mmusculus_homolog_ensembl_gene'

        # Map gene symbol to ensembl ID of query species
        query = human_dataset.search({'attributes': human_attributes})
        query = query.raw.data.decode('ascii').split('\n')[:-1]
        query = pd.DataFrame([d.split('\t') for d in query], columns=['human_ensembl_id', 'hgnc_symbol'])
        outer_bar.update(2)

        # Map ensembl IDs between two species
        from2to = human_dataset.search({'attributes': ['ensembl_gene_id', to_homolog_attribute]})
        from2to = from2to.raw.data.decode('ascii').split('\n')[:-1]
        from2to = pd.DataFrame([d.split('\t') for d in from2to], columns=['human_ensembl_id', 'mouse_ensembl_id'])
        from2to = from2to.merge(query, how='outer', on='human_ensembl_id')
        outer_bar.update()

        # Map ensembl ID to gene symbol of target species
        target = mouse_dataset.search({'attributes': mouse_attributes})
        target = target.raw.data.decode('ascii').split('\n')[:-1]
        target = pd.DataFrame([d.split('\t') for d in target], columns=['mouse_ensembl_id', 'mgi_symbol'])
        target = target.merge(from2to, how='outer', on='mouse_ensembl_id')
        outer_bar.update()

        # Gene mapper of the network
        query_genes = np.unique(net.loc[:, ['from', 'to']].astype(str))
        mapper = target.loc[target['hgnc_symbol'].isin(query_genes), ['hgnc_symbol', 'mgi_symbol']].copy()
        mapper = mapper.dropna()
        mapper = mapper.drop_duplicates()
        mapper = mapper.loc[mapper['mgi_symbol'] != '', :]

        # Process ambiguous (1-to-many) and unambiguous (1-to-1 and many-to-1) genes separately
        human_gene_value_counts = mapper.loc[:, 'hgnc_symbol'].value_counts()
        unambiguous_genes = human_gene_value_counts[human_gene_value_counts == 1].index.tolist()
        ambiguous_genes = human_gene_value_counts[human_gene_value_counts > 1].index.tolist()
        outer_bar.update()

        # Directly convert the interactions with unambiguous genes
        net_una = net.loc[net['from'].isin(unambiguous_genes) & net['to'].isin(unambiguous_genes), ['from', 'to']]
        converted_network_unambiguous = pd.DataFrame()
        mapper_una_dict = mapper.loc[mapper['hgnc_symbol'].isin(unambiguous_genes), :].set_index(['hgnc_symbol'])[
            'mgi_symbol'].to_dict()
        converted_network_unambiguous['from'] = net_una['from'].map(mapper_una_dict)
        converted_network_unambiguous['to'] = net_una['to'].map(mapper_una_dict)
        outer_bar.update()

        # Process interactions where one gene is ambiguous and another is unambiguous
        net_a = net.loc[(net['from'].isin(ambiguous_genes) & net['to'].isin(unambiguous_genes)) |
                        (net['to'].isin(ambiguous_genes) & net['from'].isin(unambiguous_genes)), ['from', 'to']]
        mapper_a = mapper.loc[mapper['hgnc_symbol'].isin(ambiguous_genes), :]
        temp_edge = []
        with tqdm(total=len(net_a),
                  desc='Converting ambiguous gene symbols',
                  leave=False,
                  miniters=1,
                  ) as pbar:
            for source, target in zip(net_a['from'], net_a['to']):
                if source in unambiguous_genes:
                    source_convert = [mapper_una_dict[source]]
                    target_convert = mapper_a[mapper_a['hgnc_symbol'] == target]['mgi_symbol'].tolist()
                else:
                    source_convert = mapper_a[mapper_a['hgnc_symbol'] == source]['mgi_symbol'].tolist()
                    target_convert = [mapper_una_dict[target]]
                temp_edge += list(product(source_convert, target_convert))
                pbar.update(1)
        converted_network_ambiguous = pd.DataFrame(temp_edge, columns=['from', 'to'])
        outer_bar.update()

    # Combine the converted network
    prior_net_converted = pd.concat([converted_network_unambiguous, converted_network_ambiguous], axis=0)
    prior_net_converted = prior_net_converted.drop_duplicates()

    print(f"The converted prior gene interaction network: "
          f"#Genes: {len(np.unique(prior_net_converted.iloc[:, [0, 1]]))}, "
          f"#Edges: {len(prior_net_converted)}")

    return prior_net_converted


# def human_bonemarrow_setty19(fpath: Optional[str] = './data_cache/human_bonemarrow_setty19.h5ad'):
#     url = 'https://zenodo.org/record/7564872/files/human_bonemarrow_setty19.h5ad'
#     adata = sc.read(fpath, backup_url=url, sparse=True, cache=True)
#     adata.var_names_make_unique()
#     return adata


def mouse_hsc_nestorowa16(fpath: Optional[str] = './data_cache/mouse_hsc_nestorowa16_v0.h5ad', version: Optional[str] = 'v0'):
    if version=='v0':
        fpath = './data_cache/mouse_hsc_nestorowa16_v0.h5ad'
        url = 'https://zenodo.org/record/8013900/files/mouse_hsc_nestorowa16_v0.h5ad'
        print('Load mouse_hsc_nestorowa16_v0.h5ad')
    elif version=='v1':
        fpath = './data_cache/mouse_hsc_nestorowa16_v1.h5ad'
        url = 'https://zenodo.org/record/8013900/files/mouse_hsc_nestorowa16_v1.h5ad'
        print('Load mouse_hsc_nestorowa16_v1.h5ad')
    else:
        print('Wrong data!')
    adata = sc.read(fpath, backup_url=url, sparse=True, cache=True)
    adata.var_names_make_unique()
    return adata


def mouse_hsc_paul15(fpath: Optional[str] = './data_cache/mouse_hsc_paul15.h5ad'):
    url = 'https://zenodo.org/record/7564872/files/mouse_hsc_paul15.h5ad'
    adata = sc.read(fpath, backup_url=url, sparse=True, cache=True)
    adata.var_names_make_unique()
    return adata



from typing import Union, Optional, List, Dict
import numpy as np
import pandas as pd
import pickle
import time
import importlib.util
import sys

from collections import OrderedDict

from sklearn.preprocessing import StandardScaler
import scanpy as sc
from anndata import AnnData

from . import config
from .validation import _validate_obsm_key, _validate_varm_key


# Used for trend computation and branch selection
PSEUDOTIME_RES = 500


class PResults(object):
    """
    Container of palantir results
    """

    def __init__(self, pseudotime, entropy, branch_probs, waypoints):
        # Initialize
        self._pseudotime = pseudotime
        self._entropy = entropy
        self._branch_probs = branch_probs
        self._branch_probs[self._branch_probs < 0.01] = 0
        self._waypoints = waypoints

    # Getters and setters
    @property
    def pseudotime(self):
        return self._pseudotime

    @property
    def branch_probs(self):
        return self._branch_probs

    @branch_probs.setter
    def branch_probs(self, branch_probs):
        self._branch_probs = branch_probs

    @property
    def entropy(self):
        return self._entropy

    @entropy.setter
    def entropy(self, entropy):
        self._entropy = entropy

    @property
    def waypoints(self):
        return self._waypoints

    @classmethod
    def load(cls, pkl_file):
        with open(pkl_file, "rb") as f:
            data = pickle.load(f)

        # Set up object
        presults = cls(
            data["_pseudotime"],
            data["_entropy"],
            data["_branch_probs"],
            data["_waypoints"],
        )
        return presults

    def save(self, pkl_file: str):
        pickle.dump(vars(self), pkl_file)


def compute_gene_trends_legacy(
    data: Union[AnnData, PResults],
    gene_exprs: Optional[pd.DataFrame] = None,
    lineages: Optional[List[str]] = None,
    n_splines: int = 4,
    spline_order: int = 2,
    n_jobs: int = -1,
    expression_key: str = "MAGIC_imputed_data",
    pseudo_time_key: str = "palantir_pseudotime",
    fate_prob_key: str = "palantir_fate_probabilities",
    gene_trend_key: str = "palantir_gene_trends",
    save_as_df: bool = None,
) -> Dict[str, Dict[str, pd.DataFrame]]:
    """
    Computes gene expression trends along pseudotemporal trajectories.

    This function calculates gene expression trends and their standard deviations
    along pseudotemporal trajectories computed by Palantir.

    Parameters
    ----------
    data : Union[AnnData, palantir.presults.PResults]
        Either a Scanpy AnnData object or a Palantir results object.
    gene_exprs : pd.DataFrame, optional
        DataFrame of gene expressions with cells as rows and genes as columns. If not provided, gene expressions
        will be fetched from `data` using `expression_key`.
    lineages : List[str], optional
        List of lineages for which to compute the trends. If None, all columns of the fate probability matrix are used. Default is None.
    n_splines : int, optional
        Number of splines to use. Must be non-negative. Default is 4.
    spline_order : int, optional
        Order of the splines to use. Must be non-negative. Default is 2.
    n_jobs : int, optional
        Number of cores to use. If -1, all available cores will be used. Default is -1.
    expression_key : str, optional
        Key to access gene expression matrix from a layer of the AnnData object. Default is 'MAGIC_imputed_data'.
        Ignored if `gene_exprs` is provided.
    pseudo_time_key : str, optional
        Key to access pseudotime from the obs attribute of the AnnData object. Default is 'palantir_pseudotime'.
    fate_prob_key : str, optional
        Key to access fate probabilities from the obsm attribute of the AnnData object. Default is 'palantir_fate_probabilities'.
    gene_trend_key : str, optional
        Key to store the computed gene trends in the varm attribute of the AnnData object. Default is 'palantir_gene_trends'.
        The trends for each lineage are stored under 'varm[gene_trend_key + "_" + lineage_name]'.
        The pseudotime points at which the trends are computed are stored in the uns attribute
        under 'uns[gene_trend_key + "_" + lineage_name + "_pseudotime"]'.
    save_as_df : bool, optional
        If True, the trends will be saved in the varm of the AnnData object as pandas DataFrame with pseudotime as column names.
        If False, the trends will be saved as numpy array and the pseudotime in uns[gene_trend_key + "_" + lineage_name + "_pseudotime"].
        The option to save as DataFrame is there due to some versions of AnnData not being able to
        write h5ad files with DataFrames in ad.varm. Default is palantir.SAVE_AS_DF=True.

    Returns
    -------
    Dict[str, Dict[str, pd.DataFrame]]
        Dictionary of gene expression trends and standard deviations for each branch.
    """
    from joblib import delayed, Parallel
    if save_as_df is None:
        save_as_df = config.SAVE_AS_DF

    # Extract palantir results from AnnData if necessary
    if isinstance(data, AnnData):
        gene_exprs = data.to_df(expression_key)
        pseudo_time = pd.Series(data.obs[pseudo_time_key], index=data.obs_names)
        fate_probs, _ = _validate_obsm_key(data, fate_prob_key)

        pr_res = PResults(pseudo_time, None, fate_probs, None)
    else:
        pr_res = data

    # Compute for all lineages if branch is not speicified
    if lineages is None:
        lineages = pr_res.branch_probs.columns

    # Results Container
    results = OrderedDict()
    for branch in lineages:
        results[branch] = OrderedDict()
        # Bins along the pseudotime
        br_cells = pr_res.branch_probs.index[pr_res.branch_probs.loc[:, branch] > 0.7]
        bins = np.linspace(0, pr_res.pseudotime[br_cells].max(), PSEUDOTIME_RES)

        # Branch results container
        results[branch]["trends"] = pd.DataFrame(0.0, index=gene_exprs.columns, columns=bins)
        results[branch]["std"] = pd.DataFrame(0.0, index=gene_exprs.columns, columns=bins)

    # Compute for each branch
    for branch in lineages:
        print(branch)
        start = time.time()

        # Branch cells and weights
        weights = pr_res.branch_probs.loc[gene_exprs.index, branch].values
        bins = np.array(results[branch]["trends"].columns)
        res = Parallel(n_jobs=n_jobs)(
            delayed(gam_fit_predict)(
                pr_res.pseudotime[gene_exprs.index].values,
                gene_exprs.loc[:, gene].values,
                weights,
                bins,
                n_splines,
                spline_order,
            )
            for gene in gene_exprs.columns
        )

        # Fill in the matrices
        for i, gene in enumerate(gene_exprs.columns):
            results[branch]["trends"].loc[gene, :] = res[i][0]
            results[branch]["std"].loc[gene, :] = res[i][1]
        if isinstance(data, AnnData):
            if save_as_df:
                df = results[branch]["trends"]
                df.columns = df.columns.astype(str)
                data.varm[gene_trend_key + "_" + branch] = df
            else:
                data.varm[gene_trend_key + "_" + branch] = results[branch]["trends"].values
                data.uns[gene_trend_key + "_" + branch + "_pseudotime"] = results[branch][
                    "trends"
                ].columns.values
        end = time.time()
        print("Time for processing {}: {} minutes".format(branch, (end - start) / 60))

    return results


def compute_gene_trends(
    ad: AnnData,
    lineages: Optional[List[str]] = None,
    masks_key: str = "branch_masks",
    expression_key: str = None,
    pseudo_time_key: str = "palantir_pseudotime",
    gene_trend_key: str = "gene_trends",
    save_as_df: bool = None,
    **kwargs,
) -> pd.DataFrame:
    """
    Compute gene expression trends along pseudotime in the given AnnData object.

    This function computes the gene expression trends for each branch of the
    pseudotime trajectory using mellon.FunctionEstimator. The computed gene
    trends are stored in the varm attribute of the AnnData object, with keys
    in the format '{gene_trend_key}_{branch}'. Each key maps to a 2D numpy
    array where rows correspond to genes and columns correspond to the
    pseudotime grid. The pseudotime grid for each branch is stored in the uns
    attribute of the AnnData object, with keys in the format
    '{gene_trend_key}_{branch}_pseudotime'.

    Parameters
    ----------
    ad : AnnData
        AnnData object containing the gene expression data and pseudotime.
    lineages : List[str], optional
        Subset of lineages for which to compute the trends.
        If None uses all columns of the fate probability matrix.
        Default is None.
    masks_key : str, optional
        Key to access the branch cell selection masks from obsm of the AnnData object.
        Default is 'branch_masks'.
    expression_key : str, optional
        Key to access the gene expression data in the layers of the AnnData object.
        If None, uses raw expression data in .X. Default is None.
    pseudo_time_key : str, optional
        Key to access the pseudotime values in the AnnData object. Default is 'palantir_pseudotime'.
    gene_trend_key : str, optional
        Key base to store the gene expression trends in varm of the AnnData object.
        Default is 'palantir_gene_trends'.
    save_as_df : bool, optional
        If True, the trends will be saved in the varm of the AnnData object as pandas DataFrame with pseudotime as column names.
        If False, the trends will be saved as numpy array and the pseudotime in uns[gene_trend_key + "_" + lineage_name + "_pseudotime"].
        The option to save as DataFrame is there due to some versions of AnnData not being able to
        write h5ad files with DataFrames in ad.varm. Default is palantir.SAVE_AS_DF=True.
    **kwargs
        Additional arguments to be passed to mellon.FunctionEstimator.

    Returns
    -------
    dict
        A dictionary containing gene expression trends for each branch. The keys of the dictionary
        are the branch names. The value for each branch is a sub-dictionary with a key 'trends' that
        maps to a DataFrame. The DataFrame contains the gene expression trends, indexed by gene names
        and columns representing pseudotime points.
    """
    if save_as_df is None:
        save_as_df = config.SAVE_AS_DF
    # Check the AnnData object for the necessary keys
    if pseudo_time_key not in ad.obs_keys():
        raise ValueError(f"'{pseudo_time_key}' is not found in the AnnData object's obs.")

    masks, branches = _validate_obsm_key(ad, masks_key, as_df=False)

    gene_exprs = ad.to_df(expression_key)
    pseudo_time = ad.obs[pseudo_time_key].values

    if lineages is not None:
        for lin in lineages:
            if lin not in branches:
                raise ValueError(
                    f"Lineage '{lin}' does not seem to have a selection in obsm['{masks_key}']."
                )
    else:
        lineages = branches

    # Import mellon locally to avoid JAX fork warnings in other parts of the code
    import mellon
    
    # Set the default arguments for mellon.FunctionEstimator
    mellon_args = dict(sigma=1, ls=1)

    lagacy_results = dict()
    # Compute the gene expression trends
    for i, branch in enumerate(branches):
        if branch not in lineages:
            continue
        print(branch)
        mask = masks[:, i]
        pt = pseudo_time[mask]
        pt_grid = np.linspace(pt.min(), pt.max(), PSEUDOTIME_RES)
        expr = gene_exprs.loc[mask, :]
        mellon_args["landmarks"] = pt_grid
        mellon_args.update(kwargs)
        func_est = mellon.FunctionEstimator(**mellon_args)
        result = func_est.fit_predict(pt, expr, pt_grid).T
        result = np.asarray(result)

        lagacy_results[branch] = {
            "trends": pd.DataFrame(result, columns=pt_grid, index=ad.var_names)
        }

        if save_as_df:
            ad.varm[gene_trend_key + "_" + branch] = pd.DataFrame(
                result,
                columns=pt_grid.astype(str),
                index=ad.var_names,
            )
        else:
            ad.varm[gene_trend_key + "_" + branch] = result
            ad.uns[gene_trend_key + "_" + branch + "_pseudotime"] = pt_grid

    return lagacy_results


def gam_fit_predict(x, y, weights=None, pred_x=None, n_splines=4, spline_order=2):
    """
    Function to compute individual gene trends using pyGAM

    This function requires the optional pygam package. If not installed, it will raise an
    ImportError with instructions on how to install it.

    Parameters
    ----------
    x : array-like
        Pseudotime axis
    y : array-like
        Magic imputed expression for one gene
    weights : array-like, optional
        Lineage branch weights
    pred_x : array-like, optional
        Pseudotime axis for predicted values
    n_splines : int, optional
        Number of splines to use. Must be non-negative.
    spline_order : int, optional
        Order of spline to use. Must be non-negative.

    Returns
    -------
    tuple
        Predicted values and their standard deviations.

    Raises
    ------
    ImportError
        If pygam is not installed. Install with `pip install pygam` or
        `pip install palantir[gam]`.
    """
    try:
        from pygam import LinearGAM, s
    except ImportError:
        raise ImportError(
            "The pygam package is required for this function but not installed. "
            "You can install it with:\n"
            "    pip install pygam\n"
            "or:\n"
            "    pip install palantir[gam]"
        )

    # Check for known compatibility issues
    try:
        import scipy.sparse as sp

        test_matrix = sp.csr_matrix((1, 1))
        if not hasattr(test_matrix, "A"):
            import warnings

            warnings.warn(
                "Your scipy version may be incompatible with the installed pygam version. "
                "If you encounter errors, consider using an older scipy version or updating pygam."
            )
    except:
        pass  # Silently ignore any errors in the compatibility check

    # Weights
    if weights is None:
        weights = np.repeat(1.0, len(x))

    # Construct dataframe
    use_inds = np.where(weights > 0)[0]

    # GAM fit
    gam = LinearGAM(s(0, n_splines=n_splines, spline_order=spline_order)).fit(
        x[use_inds], y[use_inds], weights=weights[use_inds]
    )

    # Predict
    if pred_x is None:
        pred_x = x
    y_pred = gam.predict(pred_x)

    # Standard deviations
    p = gam.predict(x[use_inds])
    n = len(use_inds)
    sigma = np.sqrt(((y[use_inds] - p) ** 2).sum() / (n - 2))
    stds = (
        np.sqrt(1 + 1 / n + (pred_x - np.mean(x)) ** 2 / ((x - np.mean(x)) ** 2).sum()) * sigma / 2
    )

    return y_pred, stds


def _gam_fit_predict_rpy2(x, y, weights=None, pred_x=None):
    import rpy2.robjects as robjects
    from rpy2.robjects import pandas2ri, Formula
    from rpy2.robjects.packages import importr

    pandas2ri.activate()

    # Weights
    if weights is None:
        weights = np.repeat(1.0, len(x))

    # Construct dataframe
    use_inds = np.where(weights > 0)[0]
    r_df = pandas2ri.py2rpy(pd.DataFrame(np.array([x, y]).T[use_inds, :], columns=["x", "y"]))

    # Fit the model
    rgam = importr("gam")
    model = rgam.gam(Formula("y~s(x)"), data=r_df, weights=pd.Series(weights[use_inds]))

    # Predictions
    if pred_x is None:
        pred_x = x
    y_pred = np.array(
        robjects.r.predict(model, newdata=pandas2ri.py2rpy(pd.DataFrame(pred_x, columns=["x"])))
    )

    # Standard deviations
    p = np.array(
        robjects.r.predict(
            model, newdata=pandas2ri.py2rpy(pd.DataFrame(x[use_inds], columns=["x"]))
        )
    )
    n = len(use_inds)
    sigma = np.sqrt(((y[use_inds] - p) ** 2).sum() / (n - 2))
    stds = (
        np.sqrt(1 + 1 / n + (pred_x - np.mean(x)) ** 2 / ((x - np.mean(x)) ** 2).sum()) * sigma / 2
    )

    return y_pred, stds


def cluster_gene_trends(
    data: Union[AnnData, pd.DataFrame],
    branch_name: str,
    genes: Optional[List[str]] = None,
    gene_trend_key: Optional[str] = "gene_trends",
    n_neighbors: int = 150,
    **kwargs,
) -> pd.Series:
    """
    Cluster gene trends using the Leiden algorithm.

    This function applies the Leiden clustering algorithm to gene expression trends
    along the pseudotemporal trajectory. If the input is an AnnData object, it uses
    the gene trends stored in the `varm` attribute accessed using the `gene_trend_key`.
    If the input is a DataFrame, it directly uses the input data for clustering.

    Parameters
    ----------
    data : Union[AnnData, pd.DataFrame]
        AnnData object or a DataFrame of gene expression trends.
    branch_name : str
        Name of the branch for which the gene trends are to be clustered.
    genes : list of str, optional
        List of genes to be considered for clustering. If None, all genes are considered. Default is None.
    gene_trend_key : str, optional
        Key to access gene trends in the AnnData object's varm. Default is 'palantir_gene_trends'.
    n_neighbors : int, optional
        The number of nearest neighbors to use for the k-NN graph construction. Default is 150.
    **kwargs
        Additional keyword arguments passed to `scanpy.tl.leiden`.

    Returns
    -------
    pd.Series
        A pandas series with the cluser lables for all passed genes.

    Raises
    ------
    KeyError
        If `gene_trend_key` is None when `data` is an AnnData object.
    """
    if isinstance(data, AnnData):
        if gene_trend_key is None:
            raise KeyError("Must provide a gene_trend_key when data is an AnnData object.")
        trends, pseudotimes = _validate_varm_key(data, gene_trend_key + "_" + branch_name)
    else:
        trends = data

    if genes is not None:
        trends = trends.loc[genes, :]

    # Standardize the trends
    trends = pd.DataFrame(
        StandardScaler().fit_transform(trends.T).T,
        index=trends.index,
        columns=trends.columns,
    )

    gt_ad = AnnData(trends.values, dtype=np.float32)
    sc.pp.neighbors(gt_ad, n_neighbors=n_neighbors, use_rep="X")
    
    # Add required kwargs for leiden with igraph backend to avoid FutureWarning
    leiden_kwargs = {'flavor': 'igraph', 'n_iterations': 2, 'directed': False}
    # Override with user-provided kwargs if any
    leiden_kwargs.update(kwargs)
    
    sc.tl.leiden(gt_ad, **leiden_kwargs)

    communities = pd.Series(gt_ad.obs["leiden"].values, index=trends.index)

    if isinstance(data, AnnData):
        col_name = gene_trend_key + "_clusters"
        if genes is None:
            data.var[col_name] = communities
        else:
            data.var[col_name] = communities
            data.var[col_name] = data.var[col_name].astype("category")

    return communities


def select_branch_cells(
    ad: AnnData,
    pseudo_time_key: str = "palantir_pseudotime",
    fate_prob_key: str = "palantir_fate_probabilities",
    q: float = 1e-2,
    eps: float = 1e-2,
    masks_key: str = "branch_masks",
    save_as_df: bool = None,
):
    """
    Selects cells along specific branches of pseudotime ordering.

    This function identifies cells that are most likely to follow a certain lineage or "fate" by looking at
    their pseudotime order and fate probabilities. These cells are expected to be along the path of differentiation
    towards that specific fate.

    Parameters
    ----------
    ad : AnnData
        Annotated data matrix. The pseudotime and fate probabilities should be stored under the keys provided.
    pseudo_time_key : str, optional
        Key to access the pseudotime from obs of the AnnData object. Default is 'palantir_pseudotime'.
    fate_prob_key : str, optional
        Key to access the fate probabilities from obsm of the AnnData object. Default is 'palantir_fate_probabilities'.
    q : float, optional
        Quantile used to determine the threshold for the fate probability.
        This parameter should be between 0 and 1. Default is 1e-2.
    eps : float, optional
        A small constant substracted from the fate probability threshold. Default is 1e-2.
    masks_key : str, optional
        Key under which the resulting branch cell selection masks are stored in the obsm of the AnnData object.
        Default is 'branch_masks'.
    save_as_df : bool, optional
        If True, the masks will be saved in obsm of the AnnData object as pandas DataFrame. If False, the masks will be
        saved as numpy array. The option to save as numpy array is there due to some versions of AnnData not being able to
        write h5ad files with DataFrames in ad.obsm. Default is palantir.SAVE_AS_DF.

    Returns
    -------
    masks : np.ndarray
        An array of boolean masks that indicates whether each cell is on the path to each fate.
    """

    if save_as_df is None:
        save_as_df = config.SAVE_AS_DF

    # make sure that the necessary keys are in the AnnData object
    if pseudo_time_key not in ad.obs:
        raise KeyError(f"{pseudo_time_key} not found in ad.obs")

    fate_probs, fate_names = _validate_obsm_key(ad, fate_prob_key, as_df=False)
    pseudotime = ad.obs[pseudo_time_key].values

    fate_probs[np.isnan(fate_probs)] = 1 / fate_probs.shape[1]

    idx = np.argsort(pseudotime)
    sorted_fate_probs = fate_probs[idx, :]
    prob_thresholds = np.empty_like(fate_probs)
    n = fate_probs.shape[0]

    pseudotime_resolution = min(PSEUDOTIME_RES, n)

    step = n // pseudotime_resolution
    nsteps = n // step
    for i in range(nsteps):
        l, r = i * step, (i + 1) * step
        mprob = np.quantile(sorted_fate_probs[:r, :], 1 - q, axis=0)
        prob_thresholds[l:r, :] = mprob[None, :]
    mprob = np.quantile(sorted_fate_probs, 1 - q, axis=0)
    prob_thresholds[r:, :] = mprob[None, :]

    prob_thresholds = np.maximum.accumulate(prob_thresholds, axis=0)

    masks = np.empty_like(fate_probs).astype(bool)
    masks[idx, :] = prob_thresholds - eps < sorted_fate_probs

    if save_as_df:
        ad.obsm[masks_key] = pd.DataFrame(masks, columns=fate_names, index=ad.obs_names)
    else:
        ad.obsm[masks_key] = masks
        ad.uns[masks_key + "_columns"] = fate_names

    return masks


import pandas as pd

#try:
#    import importlib.resources as res
#    package = res.files('CEFCON')
#    TFs_human = pd.read_csv(package / 'resources/hs_hgnc_tfs_lambert2018.txt', names=['gene_symbol'])
#    TFs_mouse = pd.read_csv(package / 'resources/mm_mgi_tfs.txt', names=['gene_symbol'])
#    TFs_human_animaltfdb = pd.read_csv(package / 'resources/hs_hgnc_tfs_animaltfdb4.txt', names=['gene_symbol'])
#    TFs_mouse_animaltfdb = pd.read_csv(package / 'resources/mm_mgi_tfs_animaltfdb4.txt', names=['gene_symbol'])

#except ImportError:  # ImportError occurs when using Python < 3.9
import pkg_resources
package = 'omicverse'
TFs_human_path = pkg_resources.resource_filename(package, 'externel/CEFCON/resources/hs_hgnc_tfs_lambert2018.txt')
TFs_mouse_path = pkg_resources.resource_filename(package, 'externel/CEFCON/resources/mm_mgi_tfs.txt')
TFs_human_animaltfdb_path = pkg_resources.resource_filename(package, 'externel/CEFCON/resources/hs_hgnc_tfs_animaltfdb4.txt')
TFs_mouse_animaltfdb_path = pkg_resources.resource_filename(package, 'externel/CEFCON/resources/mm_mgi_tfs_animaltfdb4.txt')

TFs_human = pd.read_csv(TFs_human_path, names=['gene_symbol'])
TFs_mouse = pd.read_csv(TFs_mouse_path, names=['gene_symbol'])
TFs_human_animaltfdb = pd.read_csv(TFs_human_animaltfdb_path, names=['gene_symbol'])
TFs_mouse_animaltfdb = pd.read_csv(TFs_mouse_animaltfdb_path, names=['gene_symbol'])

__all__ = ['TFs_human', 'TFs_mouse', 'TFs_human_animaltfdb', 'TFs_mouse_animaltfdb']
